/* Copyright 2016, 2018 Reahl Software Services (Pty) Ltd. All rights reserved. */
/*
    This file is part of Reahl.

    Reahl is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation; version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

.js .reahl-bootstrapfileinputbutton > input {
    /* opacity, width & height are set to visually hide the element, yet keep it :visible to jquery and to selenium */
    opacity: 0.1; /* must be > 0 for selenium to see it as visible */
    width: 0.1px;
    height: 1px; /* it has to be :visible according to jquery, yet not visible to the eye, else things break! */
    position: absolute;
    z-index: -1;
    overflow: hidden;
}

label.reahl-bootstrapfileinputbutton {
    margin: 0;
}

.no-js .reahl-bootstrapfileinputbutton > span {
    display: none;
}

.no-js .reahl-bootstrapfileinput > span.form-control {
    display: none;
}

.reahl-bootstrapfileinput > span.form-control {
    white-space: nowrap;
}


/* Added for FileUploadInput (panel) */
.reahl-bootstrap-file-upload-panel ul {
    list-style-type: none;
    padding: 0;
}

.reahl-bootstrap-file-upload-li > span {
    padding-left: 1em;
}

.no-js .reahl-bootstrapfileinput > .input-group-btn {
    display: inline;
}

.no-js label.reahl-bootstrapfileinputbutton {
    background-color: inherit;
    border: none;
}

